"""
Internal. Functions to read input files. In current version, only Parquet is
supported.
"""

import pyarrow.parquet as pq

from . import Schema, Table


def read_table(file_path: str) -> Table:
    return pq.read_table(file_path).to_pydict()


def read_schema(file_path: str) -> Schema:
    pa_schema = pq.read_table(file_path).schema
    names = []
    types = []
    nullables = []
    metadatas = []
    for field in pa_schema:
        names.append(field.name)
        types.append(field.type)
        nullables.append(field.nullable)
        metadatas.append(field.metadata)
    return (
        {
            "Name": names,
            "Type": types,
            "Nullable": nullables,
            "Metadata": metadatas,
        },
        pa_schema.metadata,
    )



"""
Internal. Functions to use SQL on Tables.
"""

import sqlite3

import pyarrow as pa

from . import Schema, Table


TYPES_MAPPING = {
    pa.bool_(): "INTEGER",
    pa.int8(): "INTEGER",
    pa.int16(): "INTEGER",
    pa.int32(): "INTEGER",
    pa.int64(): "INTEGER",
    pa.uint8(): "INTEGER",
    pa.uint16(): "INTEGER",
    pa.uint32(): "INTEGER",
    pa.uint64(): "INTEGER",
    pa.float16(): "REAL",
    pa.float32(): "REAL",
    pa.float64(): "REAL",
    pa.binary(): "BLOB",
    pa.string(): "TEXT",
}


def create_table(schema: Schema, table: Table) -> sqlite3.Connection:
    """Create an SQLite database with a single table named "data", with a
    schema derived from a given file schema, and given file content loaded
    into it. Returns an SQLite connection instance."""
    connection = _create_table(schema)
    _load_table(connection, schema, table)
    return connection


def query(connection: sqlite3.Connection, query: str) -> Table:
    table: Table = {}
    cursor = connection.execute(query)
    for row in cursor:
        for column_name in row.keys():
            table.setdefault(column_name, []).append(row[column_name])
    return table


def _create_table(schema: Schema) -> sqlite3.Connection:
    schema_table, _ = schema  # this function doesn't use the metadata
    fields = [
        (name, _sql_type(pa_type))
        for name, pa_type in zip(schema_table["Name"], schema_table["Type"])
    ]
    fields_declaration = ", ".join(
        [f"{name} {sql_type}" for name, sql_type in fields]
    )
    connection = sqlite3.connect(":memory:")
    connection.row_factory = sqlite3.Row
    connection.execute(f"CREATE TABLE data ( {fields_declaration} )")
    return connection


def _load_table(connection: sqlite3.Connection, schema: Schema, table: Table):
    schema_table, _ = schema  # this function doesn't use the metadata
    field_names = ", ".join(schema_table["Name"])
    placeholders = ", ".join("?" * len(schema_table["Name"]))
    insert_query = (
        f"INSERT INTO data ( {field_names} ) VALUES ( {placeholders } )"
    )
    connection.executemany(insert_query, zip(*table.values()))


def _sql_type(pa_type) -> str:
    return TYPES_MAPPING.get(pa_type, "TEXT")



import click

from .format import format
from .read import read_schema, read_table
from .sql import create_table, query as sql_query
from .table import truncate


def validate_non_negative_int(ctx, param, value):
    if value is not None and value < 0:
        raise click.BadParameter(
            f"value should be a zero or a positive integer, not {value}"
        )
    else:
        return value


@click.group(name="qurcol")
def main() -> None:
    pass


@main.command()
@click.argument(
    "file_path",
    nargs=1,
    required=True,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--head",
    "-h",
    type=int,
    callback=validate_non_negative_int,
    show_default=True,
    help="Number of rows to show in the head of the file",
)
@click.option(
    "--tail",
    "-t",
    type=int,
    callback=validate_non_negative_int,
    show_default=True,
    help="Number of rows to show in the tail of the file",
)
@click.option(
    "--all",
    "-a",
    "keep_all",
    is_flag=True,
    help="Do not truncate the output, keep all rows",
)
@click.option(
    "--output",
    "-o",
    "output_format",
    default="table",
    type=click.Choice(["table", "csv"], case_sensitive=False),
    help="Output format",
)
def view(file_path, head, tail, keep_all, output_format) -> None:

    # click doesn't provide means to know if the value was set by default or
    # was expliclty defined by the user; for this reason this code manages the
    # default values for head/tail/all options

    if keep_all and (head or tail):
        raise click.UsageError(
            "--head/--tail and --all are mutually exclusive"
        )

    should_truncate = not keep_all

    if should_truncate:
        if head is None and tail is None:
            # set defaults if the user didn't define any of the three options
            head = 10
            tail = 10
        elif head is None and tail is not None:
            head = 0
        elif tail is None and head is not None:
            tail = 0

    if should_truncate and head + tail == 0:
        raise click.UsageError(
            "truncated table would be empty with given head and tail values"
        )

    source_table = read_table(file_path)

    output_table, truncated_row_count = (
        truncate(source_table, head, tail)
        if should_truncate
        else (source_table, 0)
    )

    click.echo(format(output_table, output_format))
    if truncated_row_count > 0:
        click.echo(f"\n {truncated_row_count} rows are skipped")


@main.command()
@click.argument(
    "file_path",
    nargs=1,
    required=True,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--meta/--no-meta",
    "-m",
    "show_meta",
    default=False,
    help="Whether to show the schema metadata (metadata is shown as-is)",
)
@click.option(
    "--output",
    "-o",
    "output_format",
    default="table",
    type=click.Choice(["table", "csv"], case_sensitive=False),
    help="Output format",
)
def schema(file_path, show_meta, output_format) -> None:
    schema, metadata = read_schema(file_path)
    if not show_meta:
        del schema["Metadata"]
    click.echo(format(schema, output_format))
    if show_meta:
        click.echo(f"\nMetadata:\n{metadata}")


@main.command()
@click.argument(
    "file_path",
    nargs=1,
    required=True,
    type=click.Path(exists=True, dir_okay=False),
)
@click.argument("query", nargs=1, required=True, type=str)
@click.option(
    "-c",
    "inline",
    required=True,
    is_flag=True,
    help=(
        "If specified, read the SQL query from a non-optional query argument "
        "(this option is mandatory in current version)"
    ),
)
@click.option(
    "--output",
    "-o",
    "output_format",
    default="table",
    type=click.Choice(["table", "csv"], case_sensitive=False),
    help="Output format",
)
def sql(file_path, query, inline, output_format) -> None:
    schema = read_schema(file_path)
    table = read_table(file_path)
    connection = create_table(schema, table)
    query_result = sql_query(connection, query)
    click.echo(format(query_result, output_format))



"""
Internal. Functions to format the tables as text for output.
"""

import csv
import io

from tabulate import tabulate

from . import Table


def as_csv(table: Table) -> str:
    out = io.StringIO()
    writer = csv.writer(out, dialect="excel")
    writer.writerow(table.keys())
    for row in zip(*table.values()):
        writer.writerow(row)
    return out.getvalue()


def as_text_table(table: Table) -> str:
    headers = table.keys()
    return tabulate(table, headers, tablefmt="presto")  # type: ignore


_formats = {"csv": as_csv, "table": as_text_table}


def format(table: Table, format: str) -> str:
    assert format in _formats, f"unknown format {format}"
    return _formats[format](table)



"""
Internal. Functions to manipulate tables.

All functions in this module assume that input Tables are valid in a sense that
they contain same number of values in every column and consistent data types in
each column. This property is not validated in run time as it adds a
significant performance hit. Invalid tables may produce unpredictable behavior.
"""

from typing import Tuple

from . import Table


def truncate(
    table: Table, head_rows: int, tail_rows: int
) -> Tuple[Table, int]:
    """Truncates a `table` to only have at most `head_rows` in the beginning
    and `tail_rows` in the end of the resulting table, and a special row
    with `...` values in every cell to indicate that the table was truncated.
    Returns a truncated table and a number of truncated rows."""

    assert head_rows >= 0, "cannot keep a negative number of head rows"
    assert tail_rows >= 0, "cannot keep a negative number of tail rows"
    assert head_rows + tail_rows > 0, "truncated table would be empty"

    _, row_count = shape(table)

    # early exit if the table doesn't need to be truncated:
    if row_count <= head_rows + tail_rows:
        return table, 0

    result = {}
    for column_name in table:
        column = table[column_name]
        head = column[:head_rows]
        tail = column[-tail_rows:] if tail_rows > 0 else []
        result[column_name] = head + ["..."] + tail
    return result, row_count - head_rows - tail_rows


def shape(table: Table) -> Tuple[int, int]:
    column_names = list(table.keys())
    column_count = len(column_names)
    if column_count == 0:
        return 0, 0
    else:
        row_count = len(table[column_names[0]])
        return column_count, row_count



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['qurcol']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'pyarrow>=0.17.1,<0.18.0',
 'pytz>=2020.1,<2021.0',
 'tabulate>=0.8.7,<0.9.0']

entry_points = \
{'console_scripts': ['qurcol = qurcol.cli:main']}

setup_kwargs = {
    'name': 'qurcol',
    'version': '1.0.0',
    'description': "qurcol (as in 'query columnar ...') is a command line tool that enables its users to quickly explore the content of a file in a column-oriented storage format",
    'long_description': '# qurcol - view, query and convert columnar storage files from command line\n\n`qurcol` (as in "query columnar ...") is a tool that enables its users to\nquickly explore the content of a file in a column-oriented storage format (like\n[Apache Parquet](https://parquet.apache.org/), for example), using command line\nonly and without the need for more complex software like Spark or Pandas.\n\nIt allows viewing the file content, schema, querying the content with SQL and\nconverting the data to CSV.\n\nThis tool only targets the use case of a basic exploration of a file content.\nThe author believes that aforementioned Spark, Pandas, etc. should be used\ninstead in any scenario which goes beyond that.\n\n\n## Features\n\nCommand line tool to:\n\n - view a columnar file content\n - print a columnar file schema\n - convert a columnar file to a CSV file\n - run SQL queries on the data from a columnar file\n\nList of supported columnar file formats:\n\n - [Apache Parquet](https://parquet.apache.org/)\n\nUsers should be aware of the size of the source files and keep in mind that\nthe file is read in memory when being processed by this tool.\n\n\n## Status\n\nThis software is generally available. This software is intended to be used in\ncommand line by individual users. It is not intended for use in a production\nenvironment.\n\nThe software is provided "as is", without warranty of any kind, express or\nimplied, including but not limited to the warranties of merchantability,\nfitness for a particular purpose and noninfringement. In no event shall the\nauthors or copyright holders be liable for any claim, damages or other\nliability, whether in an action of contract, tort or otherwise, arising from,\nout of or in connection with the software or the use or other dealings in the\nsoftware.\n\n\n\n## Installation\n\nInstall from PyPI:\n\n    pip install [--user] qurcol\n\nAlternatively, you can download a release from the Release page in GitHub.\n\n\n## Usage\n\nBuilt-in help provides detailed usage information with examples:\n\n    qurcol --help\n\n\nFew examples are given below to demonstrate the usage. Please, refer to the\nbuilt-in help for all details though (it is not practical to keep a duplicate\nof the "help" in this README).\n\nPrint an extract (few first rows, few last rows) of the file content:\n\n    qurcol view [[--head=N] [--tail=N] | [--all]] [--output=table|csv] FILE\n\nFor example, view to the snippet of a file content:\n\n    qurcol view FILE\n\nor, export entire file in CSV format:\n\n    qurcol --all --output=csv FILE > OUTPUT_FILE\n\nPrint the file schema:\n\n    qurcol schema [--output=table|csv] FILE\n\nRun an inline SQL query on the file content:\n\n    qurcol sql -c FILE "QUERY"\n\n\n### SQL syntax\n\n`qurcol sql` command loads the file into an in-memory SQLite database. Therefore,\ndata types and SQL syntax are those of SQLite v.3. The command will attempt\nits best to map the data types from the source file to the data types available\nin SQLite, but users should be aware of the fact that SQLite data types are\noften less expressive (for example, there is no data type to represent\ndate/time information).\n\nData is loaded into a table with the name `data`.\n\nFinally, for a complete example, the following command:\n\n    qurcol sql -c --output=csv FILE "select * from data"\n\nwill have same effect as:\n\n    qurcol view --all --output=csv FILE\n\n\n## Why?\n\nHere are few sample use cases:\n\n - As a Data Engineer I want to review the schema of a file in a data\n   lake, in order to consume them accordingly in my software.\n\n - As a Data Ops Engineer I want to review the content of the sample file from\n   a data lake, in order to ensure that the data is being produced into it.\n\n - As a Data Engineer I want to query some data from a file in a data lake, in\n   order to review/confirm the properties of the data I am going to use.\n\n - As a Product Manager I want to load into a spreadsheet the `.parquet` file\n   shared with me by Data Science team, in order to review its content.\n\n\n## Not features\n\nFor any tool it is equally important to know what can and and what cannot be\ndone with it. Following potential features were considered for inclusion but\ndecided to be not in the current scope of this tool, unless a clear use case is\ndefined for them.\n\n - Conversion to other "complex" data formats (e.g. Excel), because it will add\n   more dependencies to the tool, while CSV can be imported to a spreadsheet\n   easily.\n\n - Reading data from files in other "file systems" (HDFS, S3, etc.), because\n   there exist command line tools to "dowload" data from these.\n\n - Any advanced data exploration and plotting, because there is no single way\n   to do that. You may use a combination of [Jupyter](https://jupyter.org/),\n   [Pandas](https://pandas.pydata.org/) and\n   [Seaborn](https://seaborn.pydata.org/) instead, for example.\n\n - Any advanced form of querying the data that goes beyond SQL. You may use\n   [Pandas](https://pandas.pydata.org/) or [Apache\n   Spark](https://spark.apache.org/) instead, for example.\n\n\n## Contributions\n\nContributions are very welcome. Please, feel free to submit an issue or create\na Pull Request.\n\n### Development environment\n\nThis software is written in Python 3, and has a modern development environment\nthat depends on [Poetry](https://python-poetry.org/docs/) and\n[Nox](https://nox.thea.codes/en/stable/).\n\nRun all tests:\n\n    nox -s tests\n\nOr, to quickly run tests on a single Python version only:\n\n    poetry run pytest --cov\n\nRun linters:\n\n    nox lint\n\nReformat code:\n\n    nox -s black\n\nOr, check code format without reformatting:\n\n    nox -s black -- --check .\n\nFull pre-commit check:\n\n    nox\n\nNote: Nox is configured to reuse virtualenvs by default; if you want to run Nox\nin a clean environment, add `--no-reuse-existing-virtualenvs` argument.\n\n\n### Criteria for Pull Requests\n\n - The PR should pass on CI. CI is configured to run all essential controls\n   (tests, flake8, mypy, black). You can easily run same controls in a local\n   development environment before the PR submission.\n\n - Beyond code formatting, please, try to stick to the overall code style, such\n   as the choice of variable names, code structure, etc.\n\n\n## License\n\nLicensed under the [Apache License v2.0](http://www.apache.org/licenses/).\n',
    'author': 'Timur Rubeko',
    'author_email': 'timur@rubeko.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/candidtim/qurcol',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
